#!/usr/bin/python
# get_merges.py - list all merges relevant to a LP user/team.
# Copyright (C) 2008 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import codecs
import logging
import optparse
import os
import os.path
import re
import string
import urllib
import subprocess
import sys

logging.basicConfig()

# Used to check whether a file is a proper patch of gz file
try:
    import magic
except ImportError:
    logging.error('python-magic is needed')
    sys.exit(1)

# mako template engine to generate the team report
try:
    from mako.template import Template
    from mako.runtime import Context
except ImportError:
    logging.error('python-mako is needed')
    sys.exit(1)


# default directory holding the downloaded patches and diff.gz files.
dfl_merge_dir='./merges/'

# list of components order by descending order of importance
components = ['main', 'restricted', 'universe', 'multiverse']


usage = "usage: %prog [options] LP-id"
parser = optparse.OptionParser(usage=usage)

parser.add_option('-m', '--merge-dir',
                  dest="merge_dir", default=dfl_merge_dir,
    help="Directory where merge files are downloaded and kept. Default: %s"\
         % (dfl_merge_dir))
parser.add_option('-v', '--verbose',
                  dest="verbose", default=False,
                  action="store_true", help="Verbose mode.")

(options, args) = parser.parse_args()

if options.verbose:
    logging.getLogger().setLevel(logging.DEBUG)

try:
    team=args[0]
except IndexError:
    parser.error("Need LP id")

ms=magic.open(magic.MAGIC_NONE)
ms.load()

# pattern used to extract the list of packages subscribed to by the team
p = re.compile('>([^\s]+)\sin\subuntu')

class Merge:
    def __init__(self, name, base_version, ubuntu_version,
                 debian_version, comp, outstanding):
        self.name = name
        self.base_version = base_version
        self.ubuntu_version = ubuntu_version
        self.debian_version = debian_version
        self.comp = comp
        self.outstanding = outstanding
        self.__uchlog = None

    def __repr__(self):
        return "%s: %s" % (self.__class__, self.name)

    def __cmp__(self,other):
        # a merge is more important:
        #    1. if its component is more important
        #    2. elif it is outstanding
        #    3. elif alphabetical order of package name
        if self.comp != other.comp:
            return not cmp(components.index(self.comp),
                           components.index(other.comp))
        if self.outstanding != other.outstanding:
            # True > False
            return not cmp(self.outstanding, other.outstanding)
        return cmp(self.name, other.name)

    def _get_m_dir(self):
        d = os.path.join(options.merge_dir,self.name)
        if not os.path.exists(d):
            os.makedirs(d)
        return d
    def _set_m_dir(self,d):
        raise NotImplemented
    merge_dir = property(_get_m_dir, _set_m_dir)

    def __get_file(self, filename):
        """Retrieve a file from the MoM if it's not already downloaded"""
        full_fname = "%s_%s" % (self.name,filename)
        local_fname = os.path.join(self.merge_dir, full_fname)
        if not os.path.exists(local_fname):
            logging.debug("Retrieving %s" % (local_fname))
            if self.name.startswith('lib'):
                hsh = self.name[:4]
            else:
                hsh = self.name[0]
            (p, h) = urllib.urlretrieve("http://merges.ubuntu.com/%s/%s/%s"\
                % (hsh, self.name, full_fname), filename=local_fname)
        return local_fname

    def __get_diffstats(self, version):
        """Return the output of diffstats on a specifed patch file."""
        local_fname = self.__get_file(version + '.patch')
        # Don't run diffstat if the file is not a diff file.
        # It can happen when the patch is no longer on MoM and the downloaded
        # page is actually the 404 html error page.
        if 'diff' not in ms.file(local_fname):
            return ''
        return subprocess.Popen(['diffstat', '-w', '15',
                                   '-f', '0',local_fname],
                     stdout=subprocess.PIPE).communicate()[0].decode('utf-8')

    def __get_debiandir_diffstats(self, version):
        """Return the output of diffstats of only the debian/ directory."""
        local_fname = self.__get_file(version + '.patch')
        # Don't run diffstat if the file is not a diff file.
        # It can happen when the patch is no longer on MoM and the downloaded
        # page is actually the 404 html error page.
        if 'diff' not in ms.file(local_fname):
            return ''
        p1 = subprocess.Popen(['filterdiff','-z','-i',
                                          '*/debian/*', local_fname],
                                      stdout=subprocess.PIPE)
        return subprocess.Popen(['diffstat', '-w', '15',
                                   '-f', '0',local_fname],
                                stdin = p1.stdout,
                     stdout=subprocess.PIPE).communicate()[0].decode('utf-8')

    def _get_udstats(self):
        return self.__get_diffstats(self.ubuntu_version)
    def _set_udstats(self):
        raise NotImplemented
    udstats = property(_get_udstats, _set_udstats)

    def _get_short_udstats(self):
        s = self.__get_diffstats(self.ubuntu_version)
        try:
            return s.split('\n')[-2]
        except IndexError:
            return ''
    def _set_short_udstats(self):
        raise NotImplemented
    short_udstats = property(_get_short_udstats, _set_short_udstats)

    def _get_ddstats(self):
        return self.__get_diffstats(self.debian_version)
    def _set_ddstats(self):
        raise NotImplemented
    ddstats = property(_get_ddstats, _set_ddstats)

    def _get_udebdirstats(self):
        return self.__get_debiandir_diffstats(self.ubuntu_version)
    def _set_udebdirstats(self):
        raise NotImplemented
    udebdirstats = property(_get_udebdirstats, _set_udebdirstats)

    def _get_debdirdstats(self):
        return self.__get_debiandir_diffstats(self.debian_version)
    def _set_ddebdirstats(self):
        raise NotImplemented
    ddebdirstats = property(_get_debdirdstats, _set_ddebdirstats)

    def __get_changelog(self, version):
        """Return all the changelog entries as a string between 
        base_version and version."""
        local_fname = self.__get_file(version.split(':')[-1] + '.diff.gz')
        # Don't run filterdiff if the file is not a gzip file.
        # It can happen when the patch is no longer on MoM and the downloaded
        # page is actually the 404 html error page.
        if 'gzip' not in ms.file(local_fname):
            return ''
        changelog = subprocess.Popen(['filterdiff','-z','-i',
                                          '*/debian/changelog', local_fname],
                    stdout=subprocess.PIPE).communicate()[0].decode('utf-8')
        chlog_fname = os.path.join(self.merge_dir, version + '.changelog')
        chlog_fh = codecs.open(chlog_fname,'w','utf-8')
        for l in changelog.split('\n')[3:]:
            chlog_fh.write(l[1:] + '\n')
        chlog_fh.close()
        ret = u''
        for l in subprocess.Popen(['dpkg-parsechangelog',
                                      "-v%s" % (self.base_version),
                                      "-l%s" % (chlog_fname)],
                                      stdout=subprocess.PIPE)\
                           .communicate()[0].decode('utf-8').split(u'\n'):
            if not l.startswith(u'Closes'):
                ret += l + u"\n"
        return ret

    def _get_uchlog(self):
        if self.__uchlog is None:
            self.__uchlog = self.__get_changelog(self.ubuntu_version)
        return self.__uchlog
    def _set_uchlog(self):
        raise NotImplemented
    uchlog = property(_get_uchlog,_set_uchlog)

    def _get_dchlog(self):
        return self.__get_changelog(self.debian_version)
    def _set_dchlog(self):
        raise NotImplemented
    dchlog = property(_get_dchlog,_set_dchlog)


# Gather list of pkgs for the team:
#   first from launchpad, then extra_pkgs from the config file
pkgs = set([])

(f_name, h) = urllib.urlretrieve("https://bugs.launchpad.net/~%s/+packagebugs"\
        % team)
for l in open(f_name):
    m = p.search(l)
    if m:
        pkgs.add(m.group(1))
logging.debug("LP pkgs: %s" % (pkgs))
if os.path.exists('%s.list' % (team)):
    pkgs = pkgs.union(set(map(string.strip,
                              open('%s.list' % (team)).readlines())))
logging.debug("Additional pkgs: %s" % (pkgs))
if os.path.exists('%s.blacklist' % (team)):
    blacklisted_pkgs = set(map(string.strip,
                               open('%s.blacklist' % (team)).readlines()))
    logging.debug("Blacklisted pkgs: %s" % (blacklisted_pkgs))
    pkgs.difference_update(blacklisted_pkgs)
logging.debug("Final pkgs: %s" % (pkgs))

# Get the list of merge for each component
merges = []
for c in components:
    (f_name, h) = urllib.urlretrieve("http://merges.ubuntu.com/tomerge-%s" % c)
    if h.has_key('ETag'):
        for l in open(f_name):
            (pkg, prio, b, u, d, up ) = \
                    l.split(',')[0].split(' ', 5)
            o = l.split(',')[2] == 'outstanding'
            if pkg in pkgs:
                merges.append(Merge(name=pkg, base_version=b,
                                    ubuntu_version=u, debian_version=d,
                                    comp=c, outstanding=o))
merges.sort(reverse=True)
logging.debug(merges)

# Write the team's report
fh = codecs.open("%s.html" % (team),'w','utf-8')
mytemplate = Template(filename='get_merges.mako')
ctx = Context(fh, merges=merges, team=team)
mytemplate.render_context(ctx)
fh.close()
